/*
Frequency synth for the Codan 7727 radio.
It is a work in progtess. 20210807
 */

#define Version "Codan7727_VFO_15C"

// comment out line below for LCD
#define USE_OLED

#include "si5351.h"
#include "Wire.h"


// synthesiser default settings on start up.
unsigned long frequency =   10000000;          // 4 bytes required for frequency
unsigned long frequency0 =  10000000;          // 4 bytes required for frequency

//unsigned long frequency1 =  1810000;          // 4 bytes required for frequency
unsigned long frequency1 =  3630000;          // 4 bytes required for frequency  Bodge to make it go until coils fixed

unsigned long frequency2 =  7000000;          // 4 bytes required for frequency
unsigned long frequency3 =  7150000;          // 4 bytes required for frequency
unsigned long frequency4 = 10125000;          // 4 bytes required for frequency
unsigned long frequency5 = 14100000;          // 4 bytes required for frequency

unsigned long step_size = 1000;             // default tuning increment (Hz)
unsigned long offset = 1650000;             // IF offset for clockgen when driving Kestral/Hawk RFDS Radio
static signed char ham_band_index = 2;      // start on 0hz

unsigned long LSB =  1647000;           // carrier mixer for LSB
unsigned long USB =  1650000;           // carrier mixer for USB
unsigned long BFO =  LSB;

int32_t cal_factor = 14300; //  eg. 16360 for my board, found from Si5351 calibrator sketch.
byte updateFlag=HIGH;
byte MuteFlag=HIGH;

long encoder;                               // encoder direction: CCW=-1; NC=0; CW=1
Si5351 clockgen;                            // initalize clockgen

// encoder definitions
#define ENCODER_A   A0
#define ENCODER_B   A1
// A4 = SDA
// A5 = SCL
 
#define Key_1          2
#define Key_2          3    
#define Key_3          4                     
#define ENCODER_BUTTON 5    // selects tuning decade
#define MuteIn         6   
                        
// ********************* PortB0-PB5 used to select the band.
#define band1         7    //  PB0
#define band2         8    //  PB1
#define band3         9    //  PB2
#define band4        10    //  PB3
#define band5        11    //  PB4
#define band6        12    //  PB5

#define ENCODER_INPUT_PORT PINC

int sideband      = 0;   
int last_sideband = 1;
int temp          = 0;
int key           = 0;

#ifdef USE_OLED
#include "oled_ui.h"
#else
#include "lcd_ui.h"
#endif

/* HF Ham bands
Band        UK Allocation          USA Allocation    Australia Allocation
(metres)          Mhz                   Mhz                 Mhz
 160         1.810 -  2.000        1.800 -  2.000        1.800 -  1.875
  80         3.500 -  3.800        3.500 -  4.000        3.500 -  3.800
  40         7.000 -  7.200        7.000 -  7.300        7.000 -  7.300
  30        10.100 - 10.150       10.100 - 10.150       10.100 - 10.150
  20        14.100 - 14.350       14.100 - 14.350       14.100 - 14.350
  17        18.068 - 18.168       18.068 - 18.168       18.068 - 18.168
  15        21.000 - 21.450       21.000 - 21.450       21.000 - 21.450 
  12        24.890 - 24.990       24.890 - 24.990       24.890 - 24.990
  10        28.000 - 29.700       28.000 - 29.700       28.000 - 29.700 
 */
 
void setup() { 
  setup_display();
  // Start serial and initialize the Si5351
  Serial.begin(9600);
  clockgen.init(SI5351_CRYSTAL_LOAD_8PF, 0, 0);
  clockgen.set_correction(cal_factor, SI5351_PLL_INPUT_XO); // <<<<<<<this is found from Si5351 calibrator code.

  // configure rotary encoder & push-button
  pinMode(ENCODER_A, INPUT_PULLUP);     // encoder leads
  pinMode(ENCODER_B, INPUT_PULLUP);  
  pinMode(ENCODER_BUTTON, INPUT_PULLUP);
  pinMode(MuteIn, INPUT_PULLUP);
  pinMode(Key_1,  INPUT_PULLUP);    // 
  pinMode(Key_2,  INPUT_PULLUP);  
  pinMode(Key_3,  INPUT_PULLUP);
  
  DDRB  = B111111;                // sets Arduino pins 7 to 12 as outputs.
  PORTB = B000000;                // no bands
  set_synth();
  display_freq_step();            // display frequency & step-size

}

/******************************
 main loop
*******************************/
void loop() {
  check_step_size();

if (digitalRead(MuteIn)!=MuteFlag)      // read the mute status
    { MuteFlag = digitalRead(MuteIn);
      updateFlag=HIGH;
    }

  get_button();     // read the keypad

    switch (key) {
    case 0:
    {    
    }
    break;
    
    case 1:          // toggle sideband.
      { if(last_sideband == 0)
      { BFO = USB;     // USB
      last_sideband = 1;}
      else {BFO = LSB;  // LSB
      last_sideband = 0;}
      updateFlag=HIGH;
      }
    break;

    case 2:           // 
    set_band(1);
    break;

    case 3:           // 
    break;

    case 4:           // 
    set_band(-1);
    break;

    case 5:           // 
    break;

    case 6:           // 
    break;

    case 7:           // 
    break;
    }

  // check the frequency 
  encoder = read_encoder();                       // returns +1/0/-1 (cw/nc/ccw)
  if (encoder != 0) {
    frequency = frequency + (encoder * step_size); 
     updateFlag=HIGH;
  } 

  if (updateFlag==HIGH)
   {  
    updateFlag=LOW;
    set_synth();
    display_freq_step();                          // display frequency & step-size
   }
                                                  // wait until no buttons are pressed
      while (key!=0) 
       { get_button();} 
}

/******************************
 read encoder direction
*******************************/
char read_encoder() {
  // definitions
  #define DELAY 100                               // number of polling-loops   
  static unsigned short integrator;               // encoder debounce
 
  /* the array values are determined by combining the previous and current A1,A0 
  Gray Code bit-patterns to form an array-index. +1 is assigned to all clockwise 
  (CW) indexes; -1 is assigned to all counter-clockwise (CCW) indexes; and 0 is 
  assigned to the remaining positions to indicate no movement*/   
  static char encoder[] = {0,-1,1,0,1,0,0,-1,-1,0,0,1,0,1,-1,0}; // SWAPPED PHASES
  //   static char encoder[] = {0,1,-1,0,-1,0,0,1,1,0,0,-1,0,-1,1,0};
  static unsigned char encoder_index = 0;
  
  static long count = 0;                         // CW rotation = +1; CCW rotation = -1
  static long this_reading;
  static long last_reading;
  char value;                                    // return value

  /* check for encoder rotation. The absolute "count" value changes as the encoder rotates.
  Typically there are four states (counts) per click, but often more due to contact bounce.
  Bounce doesn't matter so long as the "count" decreases" for CCW rotation; remains constant 
  when there is no rotation; and increases with CW rotation. Bounce can be completely 
  eliminated if we look for the "detent" code and integrate as shown below. */  
  encoder_index <<= 2;                           // shift previous bit-pattern left two places
  encoder_index |= (ENCODER_INPUT_PORT & 0x03);  // get current bit-pattern and combine with previous
  count += encoder[( encoder_index & 0x0f )];     
  
  // debounce encoder
  if ((ENCODER_INPUT_PORT & 0x03) == 3) {        // encoder on an indent (3 decimal == 00000011 binary)
    if (integrator < DELAY) {                    
      integrator++;
    } 
  } else {                                       // encoder not resting on indent
     if (integrator > 0) {
       integrator--;
     }
  }
  
  // prepare the "return" value
  if (integrator >= DELAY) {                     // encoder deemed to be stationary
    this_reading = count;
    if (this_reading < last_reading) value = -1;
    if (this_reading == last_reading) value = 0;
    if (this_reading > last_reading) value = 1;
    last_reading = this_reading;                 // both readings now hold current "count"
    integrator = 0;                              // reset integrator
  } else {
    value = 0;
  }
  return(value);
}
  
/******************************
 check frequency band
*******************************/
void set_band(int inc) 
{              
  ham_band_index = (ham_band_index+inc);
  ham_band_index = constrain(ham_band_index, 0, 5);
  long ham_band[] = {frequency0,frequency1,frequency2,frequency3,frequency4,frequency5};
  frequency = ham_band[ham_band_index];     // change frequency band      
  updateFlag=HIGH;
}

/******************************
 check the step size
*******************************/
void check_step_size() {
  // delay definitions
  #define DELAY 3000                             // DELAY = debounce(mS)*loops/second
  #define LONG_PUSH 20000                        // number of loops considered "long"
  
  // variables
  static boolean transition = false;             // valid during button push
  static unsigned short integrator;              // varies from 0 to MAXIMUM
  static unsigned short push_duration;           // determines step size/direction
  
  // look for first high-to-low transition
  if (!transition && (digitalRead(ENCODER_BUTTON) == LOW)) {
    transition = true;
    integrator = 0;
    push_duration = 0;
  }
  
  // look for last low-to-high transition
  if (transition) {
    if (push_duration < (LONG_PUSH*2)) {         // track push duration
      push_duration++;
    }
    if (digitalRead(ENCODER_BUTTON) == HIGH) {   // ENCODER_BUTTON is HIGH
      if (integrator < DELAY) {        
        integrator++;
      } else {                                   // ENCODER_BUTTON is LOW
        if (integrator > 0) {
    integrator--;
  }
      }
    }
   
    // update the step_size
    if (integrator >= DELAY) { 
      transition = false;
      if (push_duration > LONG_PUSH) {          // LONG PUSH: decrease step size
        step_size = step_size/10;
        if (step_size < 10) {                    //
          step_size = 100000;                   // wrap around    
        }        
      } else {                                  // SHORT PUSH: increase step size
        step_size = step_size*10;
        if (step_size > 100000) {
          step_size = 10;                       // wrap around 
        }         
      }
      
      display_freq_step();                      // update step-size     
    }
  }
}


/*******************************
set the Si5351 CLK0 frequency
 ******************************/
void set_synth()
{
switch (ham_band_index) {       // limit to SSB part of bands

  case 0:
  frequency = constrain(frequency, 1000000, 30000000); // set to full coverage for a test.
  frequency0 = frequency;
  PORTB = B000000;    // select the coils and filters None for this range!!!!
  break;

  case 1:
  frequency = constrain(frequency, 1840000, 1875000);
  frequency1 = frequency;
  PORTB = B10000;    // select the coils and filters 
  break;

  case 2:
  frequency = constrain(frequency, 3525000, 3700000);
  frequency2 = frequency;
  PORTB = B01000;    // select the coils and filters 
  break; 

  case 3:
  frequency = constrain(frequency, 7040000, 7300000);
  frequency3 = frequency;
  PORTB = B00100;    // select the coils and filters 
  break;

  case 4:
  frequency = constrain(frequency, 10115000, 10140000);
  frequency4 = frequency;
  PORTB = B00010;    // select the coils and filters
  break;

  case 5:
  frequency = constrain(frequency, 14112000, 14350000);
  frequency5 = frequency;
  PORTB = B00001;    // select the coils and filters
  break;
}
      clockgen.set_freq((BFO*100), SI5351_CLK1);
      clockgen.set_freq(((frequency+offset)*100), SI5351_CLK0); // was -
}
    

void get_button ()
{       key = 0;
        key =       !digitalRead(Key_1);
        key = key + !digitalRead(Key_2)*2;
        key = key + !digitalRead(Key_3)*4; 

  //    display_keypress(); // debug info
}
